
/****************************************************************/
/*                                                              */
/*  Practica de Arquitectures Especialitzades                   */
/*                                                              */
/*  Execucio paral.lela de l'algorisme d'ordenacio de vectors   */
/*  basada en els estandards de MPI                             */
/*                                                              */
/*  Realitzada per:  Francisco Jose Aguilar Celdran             */
/*                   Sergio Blanco Cuaresma                     */
/****************************************************************/ 


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <mpi.h>

#define N 200000
#define TAM_BUFFER 50000

int valors[N];

/* --------------------------------------------------------------------------------------- */
/* fusio_subvectors_ordenats : funcio que fusiona ordenadament dos subvectors previament   */
                               ordenats ascendentment, retornant un punter al nou vector.  */
/* --------------------------------------------------------------------------------------- */

int *fusio_subvectors_ordenats(int *vector_a, int *vector_b, int tam_a, int tam_b)
{
   int *ordenat;
   int fi_vectors = 0;
   int pos_a = 0;
   int pos_b = 0,i;
   int comptador = 0;
   //int menor_a, menor_b;

   ordenat = (int *) malloc(sizeof(int) * (tam_a + tam_b));
   while (!fi_vectors) {
      
      while ((pos_a < tam_a) && (vector_a[pos_a] <= vector_b[pos_b])) {
         ordenat[comptador] = vector_a[pos_a];
         comptador++;
         pos_a++;

      }
      while ((pos_b < tam_b) && (vector_b[pos_b] <= vector_a[pos_a])) {
         ordenat[comptador] = vector_b[pos_b];
         comptador++;
         pos_b++;

      }
      if (pos_a == tam_a) {
         while (pos_b < tam_b) {
            ordenat[comptador] = vector_b[pos_b];
            comptador++;
            pos_b++;

         }
      }
      else if (pos_b == tam_b) {
         while (pos_a < tam_a) {
            ordenat[comptador] = vector_a[pos_a];
            comptador++;
            pos_a++;

         }         
      }
      if (comptador == (tam_a + tam_b)) {
         fi_vectors = 1;
      }
   }
    for (i = 0; i < 9; i++){

      }
   return ordenat;
}

/* -------------------------------------------------------------------------------- */
/* merge_sort : algorisme recursiu de fusio de n subvectors, sent n potencia de 2,  */
/*              desenvolupant una estructura arboria en la que cada dos subvector   */
/*              s'obte a partir de la fusio dels seus dos subvectors predecessors.  */
/* -------------------------------------------------------------------------------- */

int *merge_sort(int **subvector, int num_processos, int *tam)
{
   int i, comptador;
   
   if (num_processos == 1) {
      return subvector[0];
   }
   else {
      comptador = 0;
      for (i = 0; i < num_processos; i += 2) {
         subvector[comptador] = fusio_subvectors_ordenats(&subvector[i][0],&subvector[i+1][0],tam[i],tam[i+1]);
         tam[comptador] = tam[i] + tam[i + 1];
         comptador++;
      }
      return merge_sort(&subvector[0],num_processos / 2,tam);
   }
}

/* --------------------------------------------------------------------------- */
/* Programa principal : Algorisme d'ordenacio parallelitzat en MPI per a un   */
                        Multiprocessador de Memoria Distribuida.               */
/* --------------------------------------------------------------------------- */

int main(int num_args, char *args[])
{
   int i, j, min_index, minim;
   int rank_proces;
   int num_processos;
   int buffer_proces[N];
   int tam_particio;
   int tam, tag = 0;
   double sum = 0;
   FILE *sal;
   MPI_Status *estat;
   int **subvector;
   int *tamanys, *resultat;
   int iteracions_totals;
   int iteracions_restants;
   int opcio_execucio;
   int index_actual;
   MPI_Request req;
   int a[5],b[4];
   int *c;
   
   // Inicialitzar MPI
   MPI_Init(&num_args,&args);
   
   // Obtenir el rank del proces actual
   MPI_Comm_rank(MPI_COMM_WORLD,&rank_proces);
   
   // Obtenir el numero total de processos
   MPI_Comm_size(MPI_COMM_WORLD,&num_processos);

   if (rank_proces == 0) {
      for(i = 0; i < N; i++) {
         valors[i]= rand();
      }
   }

   // Si hi ha dos o mes processos: execucio parallela de l'algorisme, mitjanant pas de missatges de MPI
   if (num_processos > 1) 
   {
      // En cas de no ser exacta la divisio de les N iteracions entre el numero de processos,
      // les iteracions que manquen s'afegiran al tamany de particio de l'ultim proces

      tam_particio = N / num_processos;
      iteracions_totals = tam_particio * num_processos;
      if (iteracions_totals < N) {
         iteracions_restants = N - iteracions_totals;
      }
   
      tamanys = (int *) malloc((num_processos) * sizeof(int));
      for (i = 0; i < num_processos - 1; i++) {
         tamanys[i] = tam_particio;
      }
      if (iteracions_restants > 0) {
         tamanys[num_processos - 1] = tam_particio + iteracions_restants;
      }
      else {
         tamanys[num_processos - 1] = tam_particio;
      }
      if (rank_proces == 0) {
         subvector = (int **) malloc((num_processos) * sizeof(int *));
         for (i = 0; i < num_processos; i++) {
            subvector[i] = (int *) malloc((tamanys[i]) * sizeof(int));
         }
      }         
    
      if (rank_proces == 0) {
         index_actual = tamanys[0];
         for (i = 1; i < num_processos; i++) {
            MPI_Ssend(&valors[index_actual],tamanys[i],MPI_INT,i,tag,MPI_COMM_WORLD);
            index_actual += tamanys[i];
         }
      }
      else {
         MPI_Recv(&buffer_proces[0],tamanys[rank_proces],MPI_INT,0,tag,MPI_COMM_WORLD,estat);
      }
      
      // Algorisme d'ordenacio de subvectors
      // El proces Host no necessita rebre el subvector donat que es ell qui el posseeix i el
      // distribueix als demes processos Kernel, per la qual cosa el proces Host treballara
      // la ordenacio sobre el vector 'valors'

      if (rank_proces != 0) {
         for (i = 0; i < tamanys[rank_proces] - 1; i++) {
            minim = buffer_proces[i + 1];
            min_index = i + 1;
            for (j = i + 2; j < tamanys[rank_proces]; j++) {
               if (buffer_proces[j] < minim) {
                  minim = buffer_proces[j];
                  min_index = j;
               }
            }
            if (buffer_proces[i] > minim) {
               buffer_proces[min_index] = buffer_proces[i];
               buffer_proces[i] = minim;
            }
         }
      }
      else {
         for (i = 0; i < tamanys[rank_proces] - 1; i++) {
            minim = valors[i + 1];
            min_index = i + 1;
            for (j = i + 2; j < tamanys[rank_proces]; j++) {
               if (valors[j] < minim) {
                  minim = valors[j];
                  min_index = j;
               }
            }
            if (valors[i] > minim) {
               valors[min_index] = valors[i];
               valors[i] = minim;
            }
         }
      }
      
      // El proces Host reb els subvectors ordenats de tots els processos kernel i els fusiona
      // en el vector final ordenat aplicant l'algorisme merge-sort
      if (rank_proces == 0) {
         subvector[0] = valors;
         for (i = 1; i < num_processos; i++) {
            MPI_Recv(&subvector[i][0],tamanys[i],MPI_INT,i,tag,MPI_COMM_WORLD,estat);
         }
         resultat = merge_sort(&subvector[0],num_processos,tamanys);
      }
      // Els processos Kernel envien els subvectors ordenats al proces Host amb comunicacio sincrona
      else {
         MPI_Ssend(buffer_proces,tamanys[rank_proces],MPI_INT,0,tag,MPI_COMM_WORLD);
      }
         
      if (rank_proces == 0) {
         for (i=0;i<N;i++) {
            if (i%100) {
               sum += resultat[i];
            }
         }
      }
   
      // Finalitzar MPI
      MPI_Finalize();
   }

   // Nomes hi ha un proces: execucio sequencial de l'algorisme d'ordenacio
   else 
   {
      for (i = 0; i < N - 1; i++) {
         minim = valors[i + 1];
         min_index = i + 1;
         for (j = i + 2; j < N; j++) {
            if (valors[j] < minim) {
               minim = valors[j];
               min_index = j;
            }
         }
         if (valors[i] > minim) {
            valors[min_index] = valors[i];
            valors[i] = minim;
         }
      }
      for (i = 0; i < N; i++) {
         if (i%100) { 
            sum += valors[i];
         }
      }
      // Finalitzar MPI
      MPI_Finalize();
   }
   if (rank_proces == 0) {
      sal=fopen("sec.sal","w");
      assert(sal!=NULL);
         
      fprintf(sal,"%f\n",sum);
      //for(i=0;i<N;i++) fprintf(sal,"%d\n",resultat[i]);
      close(sal);
   }
   
   return(0);

}

